import argparse
import pandas as pd
import scanpy as sc
import numpy as np
import random
import math
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.preprocessing import MinMaxScaler
from scipy.sparse import csr_matrix
import matplotlib.pyplot as plt
import os

def load_data(exp_path, metadata_path, metadata_label_column, tenX=False, preprocess=False):
    """
    Load gene expression data and metadata.

    Parameters:
    - exp_path (str): Path to gene expression CSV file
    - metadata_path (str): Path to metadata CSV file
    - metadata_label_column (str): Column in metadata that contains cell labels
    - tenX (bool): If True, indicates the data is in 10X format
    - preprocess (bool): If True, preprocess the gene expression data

    Returns:
    - adata (AnnData): Loaded gene expression data
    """
    if tenX:
        adata = sc.read_10x_mtx(exp_path, var_names='gene_symbols', gex_only=True, cache=False)
    else:
        gene_expression_df = pd.read_csv(exp_path, index_col=0)
        # Ensure all data are numeric
        gene_expression_df = gene_expression_df.apply(pd.to_numeric)
        # Replace zero and negative counts with a small positive value to avoid issues in log transformation
        gene_expression_df = gene_expression_df.clip(lower=1e-6)
        adata = sc.AnnData(gene_expression_df)
    #adata.raw = adata 
    metadata_df = pd.read_csv(metadata_path)
    # Convert to AnnData
    
    adata.obs['clust_assign'] = metadata_df[metadata_label_column].values
    adata.obs['clust_assign'] = adata.obs['clust_assign'].astype(str)
    adata.obs['clust_assign'] = adata.obs['clust_assign'].astype('category')
    if preprocess==False:
        return adata
        
    else: # Preprocess
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata)
        #sc.pp.scale(adata)
        #sc.tl.pca(adata)
        #sc.pp.neighbors(adata, n_neighbors=10, n_pcs=40)
        #sc.tl.louvain(adata, resolution=0.5)
    #print(adata.X)
    print("Created adata object with ", adata.shape[0], "cells and", adata.shape[1], "features.")
    return adata


import math
import numpy as np
import random
def sigmoid(x):
    """
    Sigmoid function.

    Parameters:
    - x (float): Input value

    Returns:
    - float: Sigmoid output
    """
    return 1 / (1 + math.exp(-x))

def create_binary_vector(length, num_ones):
    """
    Create a binary vector with a specified number of 1s.

    Parameters:
    - length (int): Length of the binary vector
    - num_ones (int): Number of 1s in the binary vector

    Returns:
    - np.array: Binary vector
    """
    binary_vector = np.array([1] * num_ones + [0] * (length - num_ones))
    np.random.shuffle(binary_vector)
    return binary_vector


def get_neighbor(solution, mode):
    """
    Get a neighbor of a given binary vector.

    Parameters:
    - solution (np.array): Binary vector
    - mode (int): Mode for generating neighbor

    Returns:
    - np.array: Neighbor binary vector
    """
    neighbor = solution.copy()
    flip_index = random.randint(0, len(solution) - 1)
    neighbor[flip_index] = 1 - neighbor[flip_index]  # Flip the bit

    new_Bit = neighbor[flip_index]
    if mode == 1:
        indices = [i for i in range(len(neighbor)) if neighbor[i] == new_Bit]
        flip_index = random.choice(indices)
        neighbor[flip_index] = 1 - neighbor[flip_index]

    return neighbor


"""
Cortado package
================

A brief description of the package.
"""

__version__ = '0.1.1'
__author__ = 'Musaddiq K Lodi'
__license__ = 'MIT'

from .data import *
from .marker_genes import *
from .hill_climbing import *
from .evaluate import *
from .utils import *

# Optional: initialization code
def init():
    # Code to be executed when the package is imported
    print("Cortado package initialized!")
    pass

init()


import pandas as pd
from sklearn.metrics.pairwise import cosine_similarity
import scanpy as sc
from sklearn.preprocessing import MinMaxScaler
from scipy.sparse import csr_matrix
import numpy as np 

def calc_marker_gene_score(adata, target_cluster, n_genes, pval_threshold=0.05, use_raw=False):
    """
    Calculate marker gene scores for a given cluster.

    Parameters:
    - adata (AnnData): Gene expression data
    - target_cluster (str): Target cluster label
    - n_genes (int): Number of top marker genes to select
    - pval_threshold (float): P-value threshold for marker genes
    - use_raw (bool): If True, use raw data for marker gene selection

    Returns:
    - marker_scores (pd.DataFrame): Marker gene scores
    """
    sc.tl.rank_genes_groups(
        adata,
        groupby='clust_assign',
        method='wilcoxon',
        reference='rest',
        n_genes=n_genes,
        use_raw=use_raw,
        tie_correct=True,
        corr_method='bonferroni'
    )
    
    # Step 2: Extract log2 fold changes for the target cluster
    target_log2fc = pd.DataFrame(adata.uns['rank_genes_groups']['logfoldchanges'])[target_cluster]
    genes = pd.DataFrame(adata.uns['rank_genes_groups']['names'])[target_cluster]
    # Step 3: Calculate the mean log2 fold change for each gene across all other clusters
    log2fc_all_clusters = pd.DataFrame(adata.uns['rank_genes_groups']['logfoldchanges'])
    mean_log2fc_others = log2fc_all_clusters.drop(columns=[target_cluster]).mean(axis=1)
    
    # Step 4: Calculate the marker score
    scores_df = pd.DataFrame({
        'gene': genes,
        'log2fc_target': target_log2fc,
        'pval': adata.uns['rank_genes_groups']['pvals'][target_cluster],
        'pval_adj': adata.uns['rank_genes_groups']['pvals_adj'][target_cluster],
        'mean_log2fc_others': mean_log2fc_others,
        'marker_score': target_log2fc - mean_log2fc_others
    })
    filtered_df = scores_df[scores_df['pval_adj'] < pval_threshold].dropna()
    
    # Step 5: Sort the DataFrame by marker score in descending order
    scores_df = filtered_df.sort_values(by='marker_score', ascending=False).set_index('gene')
    # Initialize the MinMaxScaler
    scaler = MinMaxScaler()
    scaled_scores_df = scores_df
    # Fit and transform the marker_score column
    scaled_scores_df['marker_score'] = scaler.fit_transform(scores_df[['marker_score']])

    return scaled_scores_df[['marker_score']]

def gene_correlation_within_cluster(target_cluster, adata):
    """
    Calculate gene correlation within a given cluster.

    Parameters:
    - target_cluster (str): Target cluster label
    - adata (AnnData): Gene expression data

    Returns:
    - sim_scores (pd.DataFrame): Gene correlation scores
    """
    cells_in_cluster = np.where(adata.obs['clust_assign'] == target_cluster)[0]
    #print("Cells in cluster", cells_in_cluster)
    #print("Printing norm exp")
    # Subset the sparse normalized expression data to cells in the cluster of interest
    clust_exp = adata.X[cells_in_cluster, :].T  # Transpose to have genes as rows
    
    # Ensure clust_exp is a dense matrix for cosine_similarity calculation
    if isinstance(clust_exp, csr_matrix):
        clust_exp = clust_exp.toarray()
    #print(clust_exp)
    # Compute cosine similarity matrix between genes
    similarity_matrix = cosine_similarity(clust_exp)
    
    # Convert the similarity matrix to a pandas DataFrame
    similarity_df = pd.DataFrame(similarity_matrix, index=adata.var_names, columns=adata.var_names)
    
    return similarity_df



import pandas as pd
import matplotlib.pyplot as plt

def evaluate(best_solution, scores_df, true_markers_file, marker_column, plot_filepath='results_fig.png', csv_filepath='eval_results.csv', plot_title='Gene Selection Evaluation'):
    """
    Evaluate gene selection against true markers.

    Parameters:
    - best_solution (np.array): Best binary vector
    - scores_df (pd.DataFrame): Marker gene scores
    - true_markers_file (str): Path to true marker genes CSV file
    - marker_column (str): Column name in true marker CSV for markers
    - plot_filepath (str): Path to save evaluation plot
    - csv_filepath (str): Path to save evaluation results CSV
    - plot_title (str): Title for the evaluation plot

    Returns:
    - precision (float): Precision
    - recall (float): Recall
    - f1 (float): F1 score
    - accuracy (float): Accuracy
    """
    selected_indices = [i for i, x in enumerate(best_solution) if x == 1]

    # Subset the scaled_scores_df using the selected indices
    selected_genes = scores_df.iloc[selected_indices].index.tolist()
    # Load true marker genes from CSV
    true_marker_df = pd.read_csv(true_markers_file)

    # Extract the true marker genes for the given cluster label
    total_marker_genes = true_marker_df[marker_column].tolist()
    # Convert total_marker_genes to gene names assuming they match 'Gene_' format
    
    # FOR SIMULATED DATA ONLY, CHANGE THIS LATER 
    #total_marker_genes_names = [f'Gene_{gene}' for gene in total_marker_genes]

    # Convert lists to sets for calculations
    selected_genes_set = set(selected_genes)
    total_marker_genes_set = set(total_marker_genes)
    # Calculate True Positives, False Positives, False Negatives, and True Negatives
    TP = len(selected_genes_set & total_marker_genes_set)
    FP = len(selected_genes_set - total_marker_genes_set)
    FN = len(total_marker_genes_set - selected_genes_set)
    TN = 0  # TN is tricky in this context as it requires knowing all possible non-marker genes

    # Calculate Precision, Recall, F1 Score
    precision = TP / (TP + FP) if (TP + FP) > 0 else 0
    recall = TP / (TP + FN) if (TP + FN) > 0 else 0
    f1 = 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0

    # Calculate Accuracy
    accuracy = TP / (TP + FP + FN) if (TP + FP + FN) > 0 else 0

    # Print results
    print(f"Precision: {precision:.4f}")
    print(f"Recall: {recall:.4f}")
    print(f"F1 Score: {f1:.4f}")
    print(f"Accuracy: {accuracy:.4f}")

    # Save results to a CSV file
    results_df = pd.DataFrame({
        'Metric': ['Precision', 'Recall', 'F1 Score', 'Accuracy'],
        'Value': [precision, recall, f1, accuracy]
    })
    results_df.to_csv(csv_filepath, index=False)
    print(f"Evaluation metrics saved to {csv_filepath}")

    # Data for plotting
    metrics = ['Precision', 'Recall', 'F1 Score', 'Accuracy']
    values = [precision, recall, f1, accuracy]

    # Create bar plot
    plt.figure(figsize=(8, 6))
    plt.bar(metrics, values, color=['blue', 'green', 'red', 'orange'])

    # Add labels and title
    plt.ylabel('Score')
    plt.title(plot_title)
    plt.ylim(0, 1)  # Set y-axis limits to 0-1

    # Show values on bars
    for i, v in enumerate(values):
        plt.text(i, v + 0.02, f"{v:.4f}", ha='center', fontsize=12)

    # Save plot
    plt.savefig(plot_filepath)
    plt.close()
    print(f"Evaluation plot saved as {plot_filepath}")

    return precision, recall, f1, accuracy


from setuptools import setup

setup(
    name="sc_cortado",
    version="0.1.21",
    author="Musaddiq Lodi",
    author_email="lodimk2@vcu.edu",
    description="CORTADO: hill Climbing Optimization foR cell-Type specific mArker gene DiscOvery",
    install_requires=[
        "numpy",
        "matplotlib",
        "scipy",
        "pandas",
        "scikit-learn",
    ],
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/lodimk2/cortado",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)


import numpy as np
import random
import math
import matplotlib.pyplot as plt
from .utils import (
    sigmoid,
    create_binary_vector,
    get_neighbor
)

def obj(X, nGenes, lambda1, lambda2, lambda3, marker_scores, sim_scores):
    """
    Objective function for stochastic hill climbing.

    Parameters:
    - X (np.array): Binary vector
    - nGenes (int): Number of genes
    - lambda1 (float): Weight for marker gene score
    - lambda2 (float): Weight for similarity score
    - lambda3 (float): Weight for gene set size
    - marker_scores (pd.DataFrame): Marker gene scores
    - sim_scores (pd.DataFrame): Gene correlation scores

    Returns:
    - float: Objective function value
    """
    c1 = lambda1 * (np.sum([X[g] * sigmoid(marker_scores.iloc[g]['marker_score']) for g in range(nGenes)]) / nGenes)

    c2 = -2 * lambda2 * (np.sum([X[g1] * X[g2] * sigmoid(sim_scores.iloc[g1, g2])
                                 for g1 in range(nGenes - 1)
                                 for g2 in range(g1 + 1, nGenes)]) /
                         (np.sum([X[g] for g in range(nGenes)]) * (np.sum([X[g] for g in range(nGenes)]) - 1) + 1))

    c3 = -lambda3 * (np.sum([X[g] for g in range(nGenes)]) / nGenes)

    return c1 + c2 + c3

def stochastic_hill_climbing_adaptive(f, initial_solution, max_iterations, gamma, idle_limit, 
                                      how_many_neighbors, nGenes, lambda1, lambda2, lambda3, marker_scores, sim_scores, mode):
    """
    Stochastic hill climbing algorithm with adaptive exploration reduction.

    Parameters:
    - f (function): Objective function
    - initial_solution (np.array): Initial binary vector
    - max_iterations (int): Maximum number of iterations
    - gamma (float): Exploration rate
    - idle_limit (int): Idle limit before stopping
    - how_many_neighbors (int): Number of neighbors to consider
    - nGenes (int): Number of genes
    - lambda1 (float): Weight for marker gene score
    - lambda2 (float): Weight for similarity score
    - lambda3 (float): Weight for gene set size
    - marker_scores (pd.DataFrame): Marker gene scores
    - sim_scores (pd.DataFrame): Gene correlation scores
    - mode (int): Mode for generating initial solution

    Returns:
    - best_solution (np.array): Best binary vector
    - best_value (float): Best objective function value
    """
    current_solution = initial_solution
    current_value = f(current_solution, nGenes, lambda1, lambda2, lambda3, marker_scores, sim_scores)

    t = 0
    idle_steps = 0
    best_solution = current_solution
    best_value = current_value

    while t < max_iterations and idle_steps < idle_limit:

        exploration_rate = gamma ** t
        print(f'At time {t}, the best value is {best_value}, with an exploration rate of {exploration_rate}.')

        Log.append(best_value)
        t += 1

        neighbors = [get_neighbor(current_solution, mode) for _ in range(how_many_neighbors)]

        if random.uniform(0, 1) < exploration_rate:
            ind = random.choice([i for i in range(how_many_neighbors)])
            current_solution, current_value = neighbors[ind], f(neighbors[ind], nGenes, lambda1, lambda2, lambda3, marker_scores, sim_scores)
            continue

        neighbor_values = [f(neighbor, nGenes, lambda1, lambda2, lambda3, marker_scores, sim_scores) for neighbor in neighbors]

        better_neighbors = [(neighbors[i], neighbor_values[i]) for i in range(len(neighbors)) if
                            neighbor_values[i] > current_value]

        if better_neighbors:
            ind = random.choice([i for i in range(how_many_neighbors)])
            current_solution, current_value = neighbors[ind], f(neighbors[ind], nGenes, lambda1, lambda2, lambda3, marker_scores, sim_scores)

            if current_value > best_value:
                best_solution, best_value = current_solution, current_value
            idle_steps = 0
        else:
            idle_steps += 1

    return best_solution, best_value
def run_stochastic_hill_climbing(marker_scores, filtered_corr_matrix, how_many=25, max_iterations=100, gamma=0.95, 
                                 idle_limit=10, lambda1=0.7, lambda2=0.2, lambda3=0.1, mode=1, plot_filename='cost_plot.png'):
    """
    Run stochastic hill climbing algorithm.

    Parameters:
    - marker_scores (pd.DataFrame): Marker gene scores
    - filtered_corr_matrix (pd.DataFrame): Filtered gene correlation scores
    - how_many (int): Number of 1s in the binary vector
    - max_iterations (int): Maximum number of iterations
    - gamma (float): Exploration rate
    - idle_limit (int): Idle limit before stopping
    - lambda1 (float): Weight for marker gene score
    - lambda2 (float): Weight for similarity score
    - lambda3 (float): Weight for gene set size
    - mode (int): Mode for generating initial solution
    - plot_filename (str): Filename for the cost plot

    Returns:
    - best_solution (np.array): Best binary vector
    - best_value (float): Best objective function value
    """
    nGenes = len(marker_scores)
    
    # Initialize solution
    if mode == 0:
        initial_solution = np.random.randint(2, size=nGenes)
    else:
        initial_solution = create_binary_vector(nGenes, how_many)

    print("Initial solution:", initial_solution)
    
    # Initialize logging for the cost function values
    global Log
    Log = []

    # Run the hill-climbing algorithm
    best_solution, best_value = stochastic_hill_climbing_adaptive(obj, initial_solution, max_iterations, gamma, idle_limit, 
                                                                  10, nGenes, lambda1, lambda2, lambda3, marker_scores, filtered_corr_matrix, mode)

    # Plot the cost function and save the plot
    plt.plot([i for i in range(len(Log))], Log)
    plt.xlabel('Iteration')
    plt.ylabel('Cost')
    plt.title('Cost Function Over Iterations')
    plt.savefig(plot_filename)
    plt.close()

    print(f"Cost plot saved as {plot_filename}")
    print("Best Solution:", best_solution)
    print("Best Value:", best_value)

    return best_solution, best_value